package com.tzld.piaoquan.ad.engine.service.score.impl;

import com.alibaba.fastjson.JSONObject;
import com.tzld.piaoquan.ad.engine.service.score.container.AdCreativeFeatureContainer;
import com.tzld.piaoquan.ad.engine.service.score.container.PidLambdaContainer;
import com.tzld.piaoquan.ad.engine.service.score.container.PidLambdaV2Container;
import com.tzld.piaoquan.ad.engine.service.score.dto.AdPlatformBidCreativeDTO;
import com.tzld.piaoquan.ad.engine.service.score.param.BidRankRecommendRequestParam;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdItemFeature;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.UserAdFeature;
import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.commons.score.ScorerUtils;
import com.tzld.piaoquan.ad.engine.commons.util.CommonCollectionUtils;
import com.tzld.piaoquan.ad.engine.service.remote.FeatureRemoteService;
import com.tzld.piaoquan.ad.engine.service.score.RankService;
import com.tzld.piaoquan.ad.engine.service.score.convert.RequestConvert;
import com.tzld.piaoquan.ad.engine.service.score.param.RankRecommendRequestParam;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import javax.annotation.PostConstruct;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

@Service
public class RankServiceImpl implements RankService {

    private final static Logger log = LoggerFactory.getLogger(RankServiceImpl.class);
    DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyyMMdd");
    DateTimeFormatter timeFormatter = DateTimeFormatter.ofPattern("yyyyMMdd HH:mm:ss");
    @Autowired
    FeatureRemoteService featureRemoteService;
    @Autowired
    RankServiceThompsonImpl rankServiceThompson;

    @Autowired
    AdCreativeFeatureContainer adCreativeFeatureContainer;
    @Value("${ad.model.cpm.max:200}")
    Double cpmMax=200d;
    @Value("${ad.model.cpm.min:30}")
    Double cpmMin=30d;

    public AdRankItem adItemRank(RankRecommendRequestParam request){
        ScoreParam param= RequestConvert.requestConvert(request);
        LocalDateTime currentTime = LocalDateTime.now();
        int currentHour = currentTime.getHour();
        int dayOfWeek = currentTime.getDayOfWeek().getValue();
        param.getRequestContext().setHour(currentHour+"");
        param.getRequestContext().setWeek(dayOfWeek+"");
        param.getRequestContext().setRegion(request.getRegion().replace("省",""));
        param.getRequestContext().setCity(request.getCity().replace("市",""));
        param.getRequestContext().setDay(currentTime.format(dateFormatter));

        UserAdFeature userAdFeature=featureRemoteService.getUserAdFeature(request.getMid());
        if(userAdFeature==null){
            userAdFeature=new UserAdFeature();
        }


        Map<Long, AdRankItem> cache=adCreativeFeatureContainer.getAll(request.getAdIdList());
        List<AdRankItem> rankItems=Collections.emptyList();
        if(!cache.isEmpty()){
            rankItems=new LinkedList<>(cache.values());
        }
//        List<AdRankItem> rankItems=adCreativeFeatureContainer.getAll(request.getAdIdList());


        //避免recommend-feature出问题
        if(rankItems==null|| rankItems.size()==0){
            rankItems=new LinkedList<>();
            for(Long adId:request.getAdIdList()){
                AdRankItem item=new AdRankItem();
                item.setAdId(adId);
                item.setItemFeature(new AdItemFeature());
                rankItems.add(item);
            }
        }
        for(AdRankItem item:rankItems){
            item.setBid1(1d);
            item.setBid2(1d);
            item.setCpa(75d);
            item.setPidLambda(PidLambdaContainer.getPidLambda(item.getAdId()));
        }

        //兜底方案
        List<AdRankItem> rankResult=rank(param, userAdFeature, rankItems,ScorerUtils.BASE_CONF);

        if (!CollectionUtils.isEmpty(rankResult)) {
//            log.info("svc=adItemRank request={} rankResult={} dataTime={}", JSONObject.toJSONString(request),JSONObject.toJSONString(rankResult),currentTime.format(timeFormatter));
            JSONObject object=new JSONObject();
            object.put("mid",request.getMid());
            object.put("adid",rankResult.get(0).getAdId());
            object.put("type",rankResult.get(0).getScore_type());
            object.put("pctr",rankResult.get(0).getCtr());
            object.put("pcvr",rankResult.get(0).getCvr());
            object.put("score",rankResult.get(0).getScore());
            object.put("pidLambda",rankResult.get(0).getPidLambda());
            object.put("lrsamples",rankResult.get(0).getLrSampleString());
            object.put("dataTime",currentTime.format(timeFormatter));
            log.info("svc=adItemRank {}", JSONObject.toJSONString(object));
            object.remove("lrsamples");
            log.info("svc=pid_log obj={}", JSONObject.toJSONString(object));
            return rankResult.get(0);
        }else {
            //空返回值
            return new AdRankItem();
        }
    }

    @Override
    public AdPlatformBidCreativeDTO adBidRank(BidRankRecommendRequestParam request) {

        ScoreParam param= RequestConvert.requestConvert(request);

        LocalDateTime currentTime = LocalDateTime.now();
        int currentHour = currentTime.getHour();
        int dayOfWeek = currentTime.getDayOfWeek().getValue();
        param.getRequestContext().setHour(currentHour+"");
        param.getRequestContext().setWeek(dayOfWeek+"");
        param.getRequestContext().setRegion(request.getRegion().replace("省",""));
        param.getRequestContext().setCity(request.getCity().replace("市",""));
        param.getRequestContext().setDay(currentTime.format(dateFormatter));

        UserAdFeature userAdFeature=featureRemoteService.getUserAdFeature(request.getMid());
        if(userAdFeature==null){
            userAdFeature=new UserAdFeature();
        }
        Map<Long,List<AdPlatformBidCreativeDTO>> groupMap=request
                .getCreativeList()
                .stream()
                .collect(Collectors.groupingBy(creativeDTO -> creativeDTO.getCreativeId()));


        Map<Long, AdRankItem> cache=adCreativeFeatureContainer.getAll(new ArrayList<>(groupMap.keySet()));
        List<AdRankItem> rankItems=Collections.emptyList();
        if(!cache.isEmpty()){
            rankItems=new LinkedList<>(cache.values());
        }
        for(AdRankItem item:rankItems){
            try {
//                AdPlatformBidCreativeDTO dto=groupMap.get(item.getAdId()+"").get(0);
                AdPlatformBidCreativeDTO dto=groupMap.get(item.getAdId()).get(0);
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                item.setCpa(dto.getCpa());
                item.setPidLambda(PidLambdaContainer.getPidLambda(item.getAdId()));
            }catch (Exception e){
                log.error("rankItems info error itemId={}",item.getAdId());
                e.printStackTrace();
            }
        }
        List<AdRankItem> rankResult;
        if(rankItems==null|| rankItems.size()==0){
            rankItems=new LinkedList<>();
            for(AdPlatformBidCreativeDTO dto:request.getCreativeList()){
                AdRankItem item=new AdRankItem();
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                item.setCpa(dto.getCpa());
                item.setAdId(dto.getCreativeId());
                item.setItemFeature(new AdItemFeature());
                item.setPidLambda(PidLambdaContainer.getPidLambda(item.getAdId()));
                rankItems.add(item);
            }
            rankResult=rankServiceThompson.rank(param, userAdFeature, rankItems,null);
        }else {
            rankResult=rank(param, userAdFeature, rankItems,ScorerUtils.BASE_CONF);
        }

        AdRankItem topItem=rankResult.get(0);

        AdPlatformBidCreativeDTO result=AdPlatformBidCreativeDTO
                .builder()
                .build();
        BeanUtils.copyProperties(topItem,result);
        result.setCreativeId(topItem.getAdId());
        result.setPctr(topItem.getCtr());
        result.setPcvr(topItem.getCvr());
//        result.setCreativeCode(groupMap.get(topItem.getAdId()+"").get(0).getCreativeCode());
        result.setCreativeCode(groupMap.get(topItem.getAdId()).get(0).getCreativeCode());
        double realECpm=0d;
            //经验值 待定
        realECpm=topItem.getEcpm1();

        if(realECpm>cpmMax/1000d){
            realECpm=cpmMax/1000d;
        }
        if(realECpm<cpmMin/1000d){
            realECpm=cpmMin/1000d;
        }
        result.setEcpm2(realECpm);
        JSONObject object=new JSONObject();
        object.put("mid",request.getMid());
        object.put("adid",result.getCreativeId());
        object.put("creativeCode",result.getCreativeCode());
        object.put("type",topItem.getScore_type());
        object.put("pctr",topItem.getCtr());
        object.put("pcvr",topItem.getCvr());
        object.put("lrsamples",topItem.getLrSampleString());
        object.put("pidLambda",topItem.getPidLambda());
        //临时加入供pid v2使用
        object.put("realECpm",realECpm);
        object.put("creativeId",result.getCreativeId());
        object.put("cpa",topItem.getCpa());
        object.put("dataTime",currentTime.format(timeFormatter));
        log.info("svc=adBidRank {}", JSONObject.toJSONString(object));
        object.remove("lrsamples");
        log.info("svc=pid_log obj={}", JSONObject.toJSONString(object));
        return result;
    }

    @Override
    public AdPlatformBidCreativeDTO adBidRankNewPid(BidRankRecommendRequestParam request) {

        ScoreParam param= RequestConvert.requestConvert(request);
        LocalDateTime currentTime = LocalDateTime.now();
        int currentHour = currentTime.getHour();
        int dayOfWeek = currentTime.getDayOfWeek().getValue();
        param.getRequestContext().setHour(currentHour+"");
        param.getRequestContext().setWeek(dayOfWeek+"");
        param.getRequestContext().setRegion(request.getRegion().replace("省",""));
        param.getRequestContext().setCity(request.getCity().replace("市",""));
        param.getRequestContext().setDay(currentTime.format(dateFormatter));

        UserAdFeature userAdFeature=featureRemoteService.getUserAdFeature(request.getMid());
        if(userAdFeature==null){
            userAdFeature=new UserAdFeature();
        }
        Map<Long,List<AdPlatformBidCreativeDTO>> groupMap=request
                .getCreativeList()
                .stream()
                .collect(Collectors.groupingBy(creativeDTO -> creativeDTO.getCreativeId()));


        Map<Long, AdRankItem> cache=adCreativeFeatureContainer.getAll(new ArrayList<>(groupMap.keySet()));
        List<AdRankItem> rankItems=Collections.emptyList();
        if(!cache.isEmpty()){
            rankItems=new LinkedList<>(cache.values());
        }
        for(AdRankItem item:rankItems){
            try {
                AdPlatformBidCreativeDTO dto=groupMap.get(item.getAdId()).get(0);
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                item.setCpa(dto.getCpa());
                item.setPidLambda(PidLambdaV2Container.getPidLambda(item.getAdId()));
            }catch (Exception e){
                e.printStackTrace();
            }
        }
        List<AdRankItem> rankResult;
        if(rankItems==null|| rankItems.size()==0){
            rankItems=new LinkedList<>();
            for(AdPlatformBidCreativeDTO dto:request.getCreativeList()){
                AdRankItem item=new AdRankItem();
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                item.setCpa(dto.getCpa());
                item.setAdId(dto.getCreativeId());
                item.setItemFeature(new AdItemFeature());
                item.setPidLambda(PidLambdaV2Container.getPidLambda(item.getAdId()));
                rankItems.add(item);
            }
            rankResult=rankServiceThompson.rank(param, userAdFeature, rankItems,null);
        }else {
            rankResult=rank(param, userAdFeature, rankItems,ScorerUtils.BASE_CONF);
        }

        AdRankItem topItem=rankResult.get(0);

        AdPlatformBidCreativeDTO result=AdPlatformBidCreativeDTO
                .builder()
                .build();
        BeanUtils.copyProperties(topItem,result);
        result.setCreativeId(topItem.getAdId());
        result.setPctr(topItem.getCtr());
        result.setPcvr(topItem.getCvr());
        result.setCreativeCode(groupMap.get(topItem.getAdId()).get(0).getCreativeCode());
        double realECpm=0d;
        //经验值 待定
        realECpm=topItem.getEcpm1();

        if(realECpm>cpmMax/1000d){
            realECpm=cpmMax/1000d;
        }
        if(realECpm<cpmMin/1000d){
            realECpm=cpmMin/1000d;
        }
        result.setEcpm2(realECpm);
        JSONObject object=new JSONObject();
        object.put("mid",request.getMid());
        object.put("adid",result.getCreativeId());
        object.put("creativeCode",result.getCreativeCode());
        object.put("type",topItem.getScore_type());
        object.put("pctr",topItem.getCtr());
        object.put("pcvr",topItem.getCvr());
        object.put("lrsamples",topItem.getLrSampleString());
        object.put("pidLambda",topItem.getPidLambda());
        //临时加入供pid v2使用
        object.put("realECpm",realECpm);
        object.put("creativeId",result.getCreativeId());
        object.put("cpa",topItem.getCpa());
        object.put("dataTime",currentTime.format(timeFormatter));
        log.info("svc=adBidRankNewPid {}", JSONObject.toJSONString(object));
        object.remove("lrsamples");
        log.info("svc=pid_v2_log obj={}", JSONObject.toJSONString(object));
        return result;
    }

    public List<AdRankItem> rank (ScoreParam param, UserAdFeature userAdFeature, List<AdRankItem> rankItems,String configFile){
        List<AdRankItem> rankResult = ScorerUtils
                .getScorerPipeline(configFile)
                .scoring(param, userAdFeature, rankItems);
        return rankResult;
    }


}
